@@ -101,11 +101,41 @@ fn make_expected_svg_path(name: &str, vl_version: VlVersion, theme: Option<&str>
         })
 }
 
+fn make_expected_scenegraph_path(
+    name: &str,
+    vl_version: VlVersion,
+    theme: Option<&str>,
+) -> PathBuf {
+    let root_path = Path::new(env!("CARGO_MANIFEST_DIR"));
+    root_path
+        .join("tests")
+        .join("vl-specs")
+        .join("expected")
+        .join(format!("{:?}", vl_version))
+        .join(if let Some(theme) = theme {
+            format!("{}-{}.sg.json", name, theme)
+        } else {
+            format!("{}.sg.json", name)
+        })
+}
+
 fn load_expected_svg(name: &str, vl_version: VlVersion, theme: Option<&str>) -> Option<String> {
     let spec_path = make_expected_svg_path(name, vl_version, theme);
     fs::read_to_string(&spec_path).ok()
 }
 
+fn load_expected_scenegraph(
+    name: &str,
+    vl_version: VlVersion,
+    theme: Option<&str>,
+) -> Option<Value> {
+    let spec_path = make_expected_scenegraph_path(name, vl_version, theme);
+    let Some(p) = fs::read_to_string(&spec_path).ok() else {
+        return None;
+    };
+    serde_json::from_str(&p).ok()
+}
+
 fn write_failed_svg(name: &str, vl_version: VlVersion, theme: Option<&str>, img: &str) -> PathBuf {
     let root_path = Path::new(env!("CARGO_MANIFEST_DIR"));
     let failed_dir = root_path
@@ -127,6 +157,33 @@ fn write_failed_svg(name: &str, vl_version: VlVersion, theme: Option<&str>, img:
     return file_path;
 }
 
+fn write_failed_scenegraph(
+    name: &str,
+    vl_version: VlVersion,
+    theme: Option<&str>,
+    sg: &Value,
+) -> PathBuf {
+    let root_path = Path::new(env!("CARGO_MANIFEST_DIR"));
+    let failed_dir = root_path
+        .join("tests")
+        .join("vl-specs")
+        .join("failed")
+        .join(format!("{:?}", vl_version));
+
+    fs::create_dir_all(failed_dir.clone()).unwrap();
+
+    let file_path = failed_dir.join(if let Some(theme) = theme {
+        format!("{}-{}.sg.json", name, theme)
+    } else {
+        format!("{}.sg.json", name)
+    });
+
+    let mut file = fs::File::create(file_path.clone()).unwrap();
+    file.write_all(serde_json::to_string_pretty(sg).unwrap().as_bytes())
+        .unwrap();
+    return file_path;
+}
+
 fn check_svg(name: &str, vl_version: VlVersion, theme: Option<&str>, img: &str) {
     let expected = load_expected_svg(name, vl_version, theme);
     if Some(img.to_string()) != expected {
@@ -138,6 +195,17 @@ fn check_svg(name: &str, vl_version: VlVersion, theme: Option<&str>, img: &str)
     }
 }
 
+fn check_scenegraph(name: &str, vl_version: VlVersion, theme: Option<&str>, sg: &Value) {
+    let expected = load_expected_scenegraph(name, vl_version, theme);
+    if Some(sg) != expected.as_ref() {
+        let path = write_failed_scenegraph(name, vl_version, theme, sg);
+        panic!(
+            "Specs don't match for {}.sg.json. Failed image written to {:?}",
+            name, path
+        )
+    }
+}
+
 fn make_expected_png_path(name: &str, vl_version: VlVersion, theme: Option<&str>) -> PathBuf {
     let root_path = Path::new(env!("CARGO_MANIFEST_DIR"));
     root_path
@@ -224,7 +292,6 @@ mod test_vegalite_to_vega {
     fn test(
         #[values(
             VlVersion::v4_17,
-            VlVersion::v5_7,
             VlVersion::v5_8,
             VlVersion::v5_9,
             VlVersion::v5_10,
@@ -233,6 +300,7 @@ mod test_vegalite_to_vega {
             VlVersion::v5_13,
             VlVersion::v5_14,
             VlVersion::v5_15,
+            VlVersion::v5_16,
         )]
         vl_version: VlVersion,
 
@@ -259,6 +327,101 @@ mod test_vegalite_to_vega {
     fn test_marker() {} // Help IDE detect test module
 }
 
+#[rustfmt::skip]
+mod test_vegalite_to_html_no_bundle {
+    use crate::*;
+    use futures::executor::block_on;
+    use vl_convert_rs::converter::VlOpts;
+    use vl_convert_rs::VlConverter;
+
+    #[rstest]
+    fn test(
+        #[values(
+            VlVersion::v5_8,
+            VlVersion::v5_9,
+            VlVersion::v5_10,
+            VlVersion::v5_11,
+            VlVersion::v5_12,
+            VlVersion::v5_13,
+            VlVersion::v5_14,
+            VlVersion::v5_15,
+            VlVersion::v5_16,
+        )]
+        vl_version: VlVersion,
+
+        #[values("circle_binned")]
+        name: &str,
+    ) {
+        initialize();
+
+        // Load example Vega-Lite spec
+        let vl_spec = load_vl_spec(name);
+
+        // Create Vega-Lite Converter and perform conversion
+        let mut converter = VlConverter::new();
+
+        let html_result = block_on(
+            converter.vegalite_to_html(vl_spec, VlOpts{vl_version, ..Default::default()}, false)
+        ).unwrap();
+
+        // Check for expected patterns
+        assert!(html_result.starts_with("<!DOCTYPE html>"));
+        assert!(html_result.contains(&format!("cdn.jsdelivr.net/npm/vega-lite@{}", vl_version.to_semver())));
+        assert!(html_result.contains("cdn.jsdelivr.net/npm/vega@5"));
+        assert!(html_result.contains("cdn.jsdelivr.net/npm/vega-embed@6"));
+    }
+
+    #[test]
+    fn test_marker() {} // Help IDE detect test module
+}
+
+#[rustfmt::skip]
+mod test_vegalite_to_html_bundle {
+    use crate::*;
+    use futures::executor::block_on;
+    use vl_convert_rs::converter::VlOpts;
+    use vl_convert_rs::VlConverter;
+
+    #[rstest]
+    fn test(
+        #[values(
+            VlVersion::v5_8,
+            VlVersion::v5_9,
+            VlVersion::v5_10,
+            VlVersion::v5_11,
+            VlVersion::v5_12,
+            VlVersion::v5_13,
+            VlVersion::v5_14,
+            VlVersion::v5_15,
+            VlVersion::v5_16,
+        )]
+        vl_version: VlVersion,
+
+        #[values("circle_binned")]
+        name: &str,
+    ) {
+        initialize();
+
+        // Load example Vega-Lite spec
+        let vl_spec = load_vl_spec(name);
+
+        // Create Vega-Lite Converter and perform conversion
+        let mut converter = VlConverter::new();
+
+        let html_result = block_on(
+            converter.vegalite_to_html(vl_spec, VlOpts{vl_version, ..Default::default()}, true)
+        ).unwrap();
+
+        // Check for expected patterns
+        assert!(html_result.starts_with("<!DOCTYPE html>"));
+        assert!(html_result.contains(vl_version.to_semver()));
+        assert!(html_result.contains("Jeffrey Heer"));
+    }
+
+    #[test]
+    fn test_marker() {} // Help IDE detect test module
+}
+
 #[rustfmt::skip]
 mod test_svg {
     use crate::*;
@@ -294,7 +457,7 @@ mod test_svg {
         let vg_spec =
             block_on(converter.vegalite_to_vega(vl_spec.clone(), VlOpts{vl_version, ..Default::default()})).unwrap();
 
-        let svg = block_on(converter.vega_to_svg(vg_spec)).unwrap();
+        let svg = block_on(converter.vega_to_svg(vg_spec, Default::default())).unwrap();
         check_svg(name, vl_version, None, &svg);
 
         // Convert directly to svg
@@ -306,6 +469,133 @@ mod test_svg {
     fn test_marker() {} // Help IDE detect test module
 }
 
+#[rustfmt::skip]
+mod test_svg_allowed_base_url {
+    use crate::*;
+    use futures::executor::block_on;
+    use vl_convert_rs::converter::{VgOpts, VlOpts};
+    use vl_convert_rs::VlConverter;
+
+    #[rstest]
+    fn test(#[values("circle_binned")] name: &str) {
+        initialize();
+
+        let vl_version = VlVersion::v5_8;
+
+        // Load example Vega-Lite spec
+        let vl_spec = load_vl_spec(name);
+
+        // Create Vega-Lite Converter and perform conversion
+        let mut converter = VlConverter::new();
+
+        // Convert to vega first
+        let vg_spec = block_on(converter.vegalite_to_vega(
+            vl_spec.clone(),
+            VlOpts {
+                vl_version,
+                ..Default::default()
+            },
+        ))
+        .unwrap();
+
+        // Check with matching base URL
+        let allowed_base_urls = Some(vec![
+            "https://raw.githubusercontent.com/vega/vega-datasets".to_string()
+        ]);
+        let svg = block_on(converter.vega_to_svg(
+            vg_spec.clone(),
+            VgOpts {
+                allowed_base_urls: allowed_base_urls.clone(),
+                ..Default::default()
+            },
+        ))
+        .unwrap();
+        check_svg(name, vl_version, None, &svg);
+
+        // Convert directly to svg
+        let svg = block_on(converter.vegalite_to_svg(
+            vl_spec.clone(),
+            VlOpts {
+                vl_version,
+                allowed_base_urls,
+                ..Default::default()
+            },
+        ))
+        .unwrap();
+        check_svg(name, vl_version, None, &svg);
+
+        // Check for error with non-matching URL
+        let allowed_base_urls = Some(vec!["https://some-other-base".to_string()]);
+
+        let Err(result) = block_on(converter.vega_to_svg(
+            vg_spec,
+            VgOpts {
+                allowed_base_urls: allowed_base_urls.clone(),
+                ..Default::default()
+            },
+        )) else {
+            panic!("Expected error")
+        };
+        assert!(result.to_string().contains("External data url not allowed"));
+
+        let Err(result) = block_on(converter.vegalite_to_svg(
+            vl_spec,
+            VlOpts {
+                vl_version,
+                allowed_base_urls: allowed_base_urls.clone(),
+                ..Default::default()
+            },
+        )) else {
+            panic!("Expected error")
+        };
+        assert!(result.to_string().contains("External data url not allowed"));
+    }
+
+    #[test]
+    fn test_marker() {} // Help IDE detect test module
+}
+
+#[rustfmt::skip]
+mod test_scenegraph {
+    use crate::*;
+    use futures::executor::block_on;
+    use vl_convert_rs::converter::VlOpts;
+    use vl_convert_rs::VlConverter;
+
+    #[rstest]
+    fn test(
+        #[values(
+            // This one has round-trip stable numeric values
+            "no_text_in_font_metrics"
+        )]
+        name: &str,
+    ) {
+        initialize();
+
+        let vl_version = VlVersion::v5_8;
+
+        // Load example Vega-Lite spec
+        let vl_spec = load_vl_spec(name);
+
+        // Create Vega-Lite Converter and perform conversion
+        let mut converter = VlConverter::new();
+
+        // Convert to vega first
+        let vg_spec =
+            block_on(converter.vegalite_to_vega(vl_spec.clone(), VlOpts{vl_version, ..Default::default()})).unwrap();
+
+        let sg = block_on(converter.vega_to_scenegraph(vg_spec, Default::default())).unwrap();
+        check_scenegraph(name, vl_version, None, &sg);
+
+        // Convert directly to svg
+        let sg = block_on(converter.vegalite_to_scenegraph(vl_spec, VlOpts{vl_version, ..Default::default()})).unwrap();
+        check_scenegraph(name, vl_version, None, &sg);
+    }
+
+    #[test]
+    fn test_marker() {} // Help IDE detect test module
+}
+
 #[rustfmt::skip]
 mod test_png_no_theme {
     use crate::*;
@@ -324,6 +614,7 @@ mod test_png_no_theme {
         case("float_font_size", 1.0),
         case("no_text_in_font_metrics", 1.0),
         case("custom_projection", 1.0),
+        case("long_legend_label", 1.0),
     )]
     fn test(
         name: &str,
@@ -344,7 +635,7 @@ mod test_png_no_theme {
             converter.vegalite_to_vega(vl_spec.clone(), VlOpts{vl_version, ..Default::default()})
         ).unwrap();
 
-        let png_data = block_on(converter.vega_to_png(vg_spec, Some(scale), None)).unwrap();
+        let png_data = block_on(converter.vega_to_png(vg_spec, Default::default(), Some(scale), None)).unwrap();
         check_png(name, vl_version, None, png_data.as_slice());
 
         // Convert directly to png
@@ -395,7 +686,8 @@ mod test_png_theme_config {
                     vl_version,
                     theme: Some(theme.to_string()),
                     config: Some(json!({"background": BACKGROUND_COLOR})),
-                    show_warnings: false
+                    show_warnings: false,
+                    allowed_base_urls: None,
                 },
                 Some(scale),
                 None
@@ -418,7 +710,8 @@ mod test_png_theme_config {
                     vl_version,
                     theme: None,
                     config: Some(json!({"background": BACKGROUND_COLOR})),
-                    show_warnings: false
+                    show_warnings: false,
+                    allowed_base_urls: None,
                 },
                 Some(scale),
                 None
@@ -492,7 +785,7 @@ mod test_jpeg {
         let vg_spec =
             block_on(converter.vegalite_to_vega(vl_spec.clone(), VlOpts{vl_version, ..Default::default()})).unwrap();
 
-        let jpeg_bytes = block_on(converter.vega_to_jpeg(vg_spec, None, None)).unwrap();
+        let jpeg_bytes = block_on(converter.vega_to_jpeg(vg_spec, Default::default(), None, None)).unwrap();
 
         // Check for JPEG prefix
         assert_eq!(&jpeg_bytes.as_slice()[..10], b"\xff\xd8\xff\xe0\x00\x10JFIF");
